"""Run quantum program on IBM QE cloud platform."""
import signal
import time
import uuid

import requests
from requests import Session
from requests.compat import urljoin

# TODO: move to config file
_AUTH_API_URL = "https://auth.quantum-computing.ibm.com/api/users/loginWithToken"
_API_URL = "https://api.quantum-computing.ibm.com/api/"

# TODO: call to get the API version automatically
CLIENT_APPLICATION = "ibmqprovider/0.4.4"


# TODO: exclude in API document.
class IBMQ(Session):
    """Manage a session between QuTrunk and the IBMQ web API."""

    def __init__(self, **kwargs):
        """Initialize a session with the IBM QE's APIs."""
        super().__init__(**kwargs)
        self.backends = {}
        self.timeout = 5.0
        self.hooks["response"].append(
            lambda r, *i_args, **i_kwargs: r.raise_for_status()
        )

    def get_list_devices(self, verbose=False):
        """Get the list of available IBM backends with their properties.

        Args:
            verbose (bool): Print the returned dictionary if True

        Returns:
            Backends(dict) dictionary by name device, containing the qubit size 'nq',
            the coupling map 'coupling_map' as well as the device version 'version'
        """
        list_device_url = "Network/ibm-q/Groups/open/Projects/main/devices/v/1"
        argument = {"allow_redirects": True, "timeout": (self.timeout, None)}
        request = super().get(urljoin(_API_URL, list_device_url), **argument)
        # 反序列化
        r_json = request.json()
        self.backends = {}
        for obj in r_json:
            self.backends[obj["backend_name"]] = {
                "nq": obj["n_qubits"],
                "coupling_map": obj["coupling_map"],
                "version": obj["backend_version"],
            }

        if verbose:
            print("- List of IBMQ devices available:")
            print(self.backends)
        return self.backends

    def authenticate(self, token=None):
        """Authenticate with IBM's Web API.

        Args:
            token (str): IBM quantum experience user API token.
        """
        if token is None:
            raise Exception("please input the IBM QE token")
        # write token into headers
        self.headers.update({"X-Qx-Client-Application": CLIENT_APPLICATION})
        args = {
            "data": None,
            "json": {"apiToken": token},
            "timeout": (self.timeout, None),
        }
        request = super().post(_AUTH_API_URL, **args)
        r_json = request.json()
        self.params.update({"access_token": r_json["id"]})

    def is_online(self, device):
        """Device is available."""
        return device in self.backends

    def can_run_experiment(self, info=None, device=""):
        """Whether the resources satisfy to run quantum programs."""
        nb_qubit_max = self.backends[device]["nq"]
        nb_qubit_needed = info["nq"]
        return nb_qubit_needed <= nb_qubit_max, nb_qubit_max, nb_qubit_needed

    def run(self, info=None, device=""):
        """Run the quantum code to the IBMQ machine.

         Args:
             info (dict): Dictionary sent by the backend containing the code to run.
             device (str): Name of the ibm device to use.

        Returns:
            Execution Id.
        """
        # 1 get the device URL and job id
        json_step1 = {
            "data": None,
            "json": {
                "backend": {"name": device},
                "allowObjectStorage": True,
                "shareLevel": "none",
            },
            "timeout": (self.timeout, None),
        }
        request = super().post(
            urljoin(_API_URL, "Network/ibm-q/Groups/open/Projects/main/Jobs"),
            **json_step1,
        )
        request.raise_for_status()
        r_json = request.json()
        print("r_json==", r_json)
        upload_url = r_json["objectStorageInfo"]["uploadUrl"]
        execution_id = r_json["id"]

        # 2 send circuit
        n_classical_reg = info["nq"]
        n_qubits = n_classical_reg
        # instruction
        instructions = info["json"]
        maxcredit = info["maxCredits"]
        c_label = [["c", i] for i in range(n_classical_reg)]
        q_label = [["q", i] for i in range(n_qubits)]

        instruction_str = str(instructions).replace("'", '"')
        data = '{"qobj_id": "' + str(uuid.uuid4()) + '", '
        data += '"header": {"backend_name": "' + device + '", '
        data += '"backend_version": "' + self.backends[device]["version"] + '"}, '
        data += '"config": {"shots": ' + str(info["shots"]) + ", "
        data += '"max_credits": ' + str(maxcredit) + ', "memory": false, '
        data += '"parameter_binds": [], "memory_slots": ' + str(n_classical_reg)
        data += ', "n_qubits": ' + str(n_qubits) + '}, "schema_version": "1.2.0", '
        data += '"type": "QASM", "experiments": [{"config": '
        data += '{"n_qubits": ' + str(n_qubits) + ", "
        data += '"memory_slots": ' + str(n_classical_reg) + "}, "
        data += '"header": {"qubit_labels": ' + str(q_label).replace("'", '"') + ", "
        data += '"n_qubits": ' + str(n_classical_reg) + ", "
        data += '"qreg_sizes": [["q", ' + str(n_qubits) + "]], "
        data += '"clbit_labels": ' + str(c_label).replace("'", '"') + ", "
        data += '"memory_slots": ' + str(n_classical_reg) + ", "
        data += '"creg_sizes": [["c", ' + str(n_classical_reg) + "]], "
        data += (
            '"name": "circuit0", "global_phase": 0}, "instructions": '
            + instruction_str
            + "}]}"
        )

        json_step2 = {
            "data": data,
            "params": {"access_token": None},
            "timeout": (5.0, None),
        }
        super().put(upload_url, **json_step2)

        # 3 upload
        json_step3 = {"data": None, "json": None, "timeout": (self.timeout, None)}

        upload_data_url = urljoin(
            _API_URL,
            "Network/ibm-q/Groups/open/Projects/main/Jobs/"
            + str(execution_id)
            + "/jobDataUploaded",
        )
        super().post(upload_data_url, **json_step3)

        return execution_id

    def get_result(
        self, device, execution_id, num_retries=3000, interval=1, verbose=False
    ):
        """Get running results"""
        job_status_url = "Network/ibm-q/Groups/open/Projects/main/Jobs/" + execution_id
        if verbose:
            print(f"Waiting for results. [Job ID: {execution_id}]")
            original_sigint_handler = signal.getsignal(signal.SIGINT)
            print("signal=", original_sigint_handler)

            def _handle_sigint_during_get_result(*_):
                raise Exception(
                    "Interrupted. The ID of your submitted job is {}.".format(
                        execution_id
                    )
                )

            try:
                signal.signal(signal.SIGINT, _handle_sigint_during_get_result)

                for retries in range(num_retries):
                    # 1: wait job running
                    json_step5 = {
                        "allow_redirects": True,
                        "timeout": (self.timeout, None),
                    }
                    request = super().get(
                        urljoin(_API_URL, job_status_url), **json_step5
                    )
                    r_json = request.json()
                    acceptable_status = ["VALIDATING", "VALIDATED", "RUNNING"]

                    # 2 get running results
                    if r_json["status"] == "COMPLETED":
                        json_step6 = {
                            "allow_redirects": True,
                            "timeout": (self.timeout, None),
                        }
                        request = super().get(
                            urljoin(_API_URL, job_status_url + "/resultDownloadUrl"),
                            **json_step6,
                        )
                        r_json = request.json()

                        json_step7 = {
                            "allow_redirects": True,
                            "params": {"access_token": None},
                            "timeout": (self.timeout, None),
                        }
                        request = super().get(r_json["url"], **json_step7)
                        r_json = request.json()
                        result = r_json["results"][0]

                        json_step8 = {
                            "data": None,
                            "json": None,
                            "timeout": (5.0, None),
                        }
                        request = super().post(
                            urljoin(_API_URL, job_status_url + "/resultDownloaded"),
                            **json_step8,
                        )
                        r_json = request.json()  # {'terminated': True}

                        return result

                    if r_json["status"] not in acceptable_status:
                        raise Exception(
                            f"Error while running the code. Last status: {r_json['status']}."
                        )
                    time.sleep(interval)
                    if self.is_online(device) and retries % 60 == 0:
                        self.get_list_devices()
                        if not self.is_online(device):
                            raise Exception(
                                f"Device went offline. The ID of your submitted job is {execution_id}."
                            )

            finally:
                if original_sigint_handler is not None:
                    signal.signal(signal.SIGINT, original_sigint_handler)

            raise Exception(f"Timeout. The ID of your submitted job is {execution_id}.")


def show_devices(token=None, verbose=False):
    """Access the list of available devices.

    Args:
        token (str): IBM quantum experience user API token.
        verbose (bool): If True, additional information is printed.

    Returns:
        List (list) of available devices and their properties.
    """
    ibmq_session = IBMQ()
    ibmq_session.authenticate(token=token)
    return ibmq_session.get_list_devices(verbose=verbose)


def send(
    info,
    device="ibmq_qasm_simulator",
    token=None,
    shots=None,
    num_retries=3000,
    interval=1,
    verbose=False,
):
    """Send QASM through the IBM API and runs the quantum circuit.

    Args:
        info(dict): Contains representation of the circuit to run.
        device (str): name of the ibm device. Simulator chosen by default.
        token (str): IBM quantum experience user API token.
        shots (int): Number of runs of the same circuit to collect statistics.
        verbose (bool): If True, additional information is printed, such as measurement statistics. Otherwise, the
            backend simply registers one measurement result (same behavior as the project Simulator).

    Returns:
        Result (dict) form the IBMQ server.

    """
    try:
        ibmq_session = IBMQ()
        if shots is not None:
            info["shots"] = shots
        if verbose:
            print("- Authenticating...")
            if token is not None:
                print("user API token: " + token)
        ibmq_session.authenticate(token)

        # check if the device is online
        ibmq_session.get_list_devices(verbose)
        online = ibmq_session.is_online(device)
        if not online:
            print(
                "The device is offline. Use the simulator instead or try again later."
            )
            raise Exception("Device is offline.")

        # check if the device has enough qubit to run the code
        runnable, qmax, qneeded = ibmq_session.can_run_experiment(info, device)
        if not runnable:
            print(
                f"The device is too small ({qmax} qubits available) for the code "
                + f"requested({qneeded} qubits needed) Try to look for another "
                + "device with more qubits"
            )
            raise Exception("Device is too small.")
        if verbose:
            print(f"- Running code: {info}")
        execution_id = ibmq_session.run(info, device)
        if verbose:
            print("- Waiting for results...")
        res = ibmq_session.get_result(
            device,
            execution_id,
            num_retries=num_retries,
            interval=interval,
            verbose=verbose,
        )
        if verbose:
            print("- Done.")
        return res
    except requests.exceptions.HTTPError as err:
        print("- There was an error running your code:")
        print(err)
    except requests.exceptions.RequestException as err:
        print("- Looks like something is wrong with server:")
        print(err)
    except KeyError as err:
        print("- Failed to parse response:")
        print(err)
    return None



import cmath

import numpy as np

from .basicgate import BasicRotateGate
from qutrunk.circuit import Command
from qutrunk.circuit.qubit import QuBit


class Rz(BasicRotateGate):
    """Rotate a single qubit by a given angle around the Z-axis of the Bloch-sphere \
        (also known as a phase shift gate).

    Args:
        alpha: The angle to rotate.

    Example:
        .. code-block:: python

            Rz(alpha) * qr[0]
    """

    def __init__(self, alpha):
        if alpha is None:
            raise ValueError("The argument cannot be empty.")
        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "Rz"

    def __or__(self, qubit):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to apply Rz gate.

        Example:
            .. code-block:: python

                Rz(alpha) * qr[0]

        Raises:
            NotImplementedError: If the argument is not a Qubit object.
        """
        if not isinstance(qubit, QuBit):
            raise TypeError("The argument must be Qubit object.")

        targets = [qubit.index]
        cmd = Command(self, targets, rotation=[self.rotation], inverse=self.is_inverse)
        self.commit(qubit.circuit, cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, \
            reuse __or__ operator implement."""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix(
            [
                [cmath.exp(-0.5 * 1j * self.rotation), 0],
                [0, cmath.exp(0.5 * 1j * self.rotation)],
            ]
        )

    def inv(self):
        """Apply inverse gate."""
        gate = Rz(self.rotation)
        gate.is_inverse = not self.is_inverse 
        return gate

    def ctrl(self, ctrl_cnt=1):
        """Apply controlled gate.
        
        Args:
            ctrl_cnt: The number of control qubits, default: 1.
        """
        if ctrl_cnt > 1:
            raise ValueError("Rz gate do not support multiple control bits.")
        gate = CRz(self.rotation)
        gate.is_inverse = self.is_inverse
        return gate


class CRz(BasicRotateGate):
    """Control Rz gate.

    Args:
        angle: The angle of the rotation.

    Example:
        .. code-block:: python

            CRz(pi/2) * (qr[0], qr[1])
    """

    def __init__(self, angle):
        if angle is None:
            raise ValueError("The argument cannot be empty.")
        super().__init__()
        self.rotation = angle

    def __str__(self):
        return "CRz"

    def __or__(self, qubits):
        """Quantum logic gate operation.

        Args:
            qubits: qubits[0] is control qubit, qubits[1] is target qubit.

        Example:
            .. code-block:: python

                CRz(pi/2) * (qr[0], qr[1])
        """
        if not all(isinstance(qubit, QuBit) for qubit in qubits):
            raise TypeError("The argument must be Qubit object.")

        if len(qubits) != 2:
            raise ValueError(
                "Parameter error: One controlled and one target qubit are required."
            )

        self.qubits = qubits
        controls = [qubits[0].index]
        targets = [qubits[1].index]
        cmd = Command(
            self, targets, controls, inverse=self.is_inverse, rotation=[self.rotation]
        )
        self.commit(qubits[0].circuit, cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        arg = 1j * float(self.rotation) / 2
        return np.array(
            [
                [np.exp(-arg), 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, np.exp(arg), 0],
                [0, 0, 0, 1],
            ]
        )

    def inv(self):
        """Apply inverse gate"""
        gate = CRz(self.rotation)
        gate.is_inverse = not self.is_inverse
        return gate



import numpy as np

from .basicgate import BasicRotateGate
from qutrunk.circuit import Command
from qutrunk.circuit.qubit import QuBit


class Rx(BasicRotateGate):
    """Rotate a single qubit by a given angle around the X-axis of the Bloch-sphere.

    Args:
        alpha: The angle to rotate.

    Example:
        .. code-block:: python

            Rx(alpha) * qr[0]
    """

    def __init__(self, alpha):
        if alpha is None:
            raise ValueError("The argument cannot be empty.")

        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "Rx"

    def __or__(self, qubit):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to aplly Rx gate。

        Example:
            .. code-block:: python

                Rx(alpha) * qr[0]

        Raises:
            NotImplementedError: If the argument is not a Qubit object.
        """
        if not isinstance(qubit, QuBit):
            raise TypeError("The argument must be Qubit object.")

        targets = [qubit.index]
        cmd = Command(self, targets, rotation=[self.rotation], inverse=self.is_inverse)
        self.commit(qubit.circuit, cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        half_theta = float(self.rotation) / 2
        cos = np.cos(half_theta)
        sin = np.sin(half_theta)
        return np.array(
            [
                [cos, -1j * sin],
                [-1j * sin, cos],
            ]
        )

    def inv(self):
        """Apply inverse gate."""
        gate = Rx(self.rotation)
        gate.is_inverse = not self.is_inverse 
        return gate

    def ctrl(self, ctrl_cnt=1):
        """Apply controlled gate.
        
        Args:
            ctrl_cnt: The number of control qubits, default: 1.
        """
        if ctrl_cnt > 1:
            raise ValueError("Rx gate do not support multiple control bits.")
        gate = CRx(self.rotation)
        gate.is_inverse = self.is_inverse
        return gate


class CRx(BasicRotateGate):
    """Control Rx Gate.

    Args:
        angle: The angle of the rotation.

    Example:
        .. code-block:: python

            CRx(pi/2) * (qr[0], qr[1])

    """

    def __init__(self, angle):
        if angle is None:
            raise ValueError("The argument cannot be empty.")

        super().__init__()
        self.rotation = angle

    def __str__(self):
        return "CRx"

    def __or__(self, qubits):
        """Quantum logic gate operation.

        Args:
            qubits: qubits[0] is control qubit, qubits[1] is target qubit.

        Example:
            .. code-block:: python

            CRx(pi/2) * (qr[0], qr[1])

        Raises:
            NotImplementedError: If the argument is not a Qubit object.
            AttributeError: If the qubits should not be two.
        """
        if not all(isinstance(qubit, QuBit) for qubit in qubits):
            raise TypeError("The argument must be Qubit object.")

        if len(qubits) != 2:
            raise ValueError(
                "Parameter error: One controlled and one target qubit are required."
            )

        self.qubits = qubits
        controls = [qubits[0].index]
        targets = [qubits[1].index]
        cmd = Command(
            self, targets, controls, inverse=self.is_inverse, rotation=[self.rotation]
        )
        self.commit(qubits[0].circuit, cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        # TODO:define have problem.
        half_theta = float(self.rotation) / 2
        cos = np.cos(half_theta)
        isin = 1j * np.sin(half_theta)
        return np.array(
            [[cos, 0, -isin, 0], [0, 1, 0, 0], [-isin, 0, cos, 0], [0, 0, 0, 1]]
        )

    def inv(self):
        """Apply inverse gate"""
        gate = CRx(self.rotation)
        gate.is_inverse = not self.is_inverse 
        return gate


from cmath import sin, cos

import numpy as np

from .basicgate import BasicRotateGate
from qutrunk.circuit import Command
from qutrunk.circuit.qubit import QuBit


class Ryy(BasicRotateGate):
    """RotationYY gate class.

    Args:
        alpha: The angle to rotate.

    Example:
        .. code-block:: python

            Ryy(alpha) * (qr[0], qr[1])
    """

    def __init__(self, alpha):
        if alpha is None:
            raise ValueError("The argument cannot be empty.")
        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "Ryy"

    def __or__(self, qubits):
        """Quantum logic gate operation.

        Args:
            qubits: The quantum bits to apply Ryy gate.

        Example:
            .. code-block:: python

                Ryy(alpha) * (qr[0], qr[1])

        Raises:
            TypeError: If the argument is not a Qubit object.
            ValueError: If the qubits should not be two.
        """
        if not all(isinstance(qubit, QuBit) for qubit in qubits):
            raise TypeError("The argument must be Qubit object.")

        if len(qubits) != 2:
            raise ValueError("Parameter error: Two target qubits are required.")

        targets = [q.index for q in qubits]
        cmd = Command(self, targets, rotation=[self.rotation], inverse=self.is_inverse)
        self.commit(qubits[0].circuit, cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix(
            [
                [cos(0.5 * self.rotation), 0, 0, 1j * sin(0.5 * self.rotation)],
                [0, cos(0.5 * self.rotation), -1j * sin(0.5 * self.rotation), 0],
                [0, -1j * sin(0.5 * self.rotation), cos(0.5 * self.rotation), 0],
                [1j * sin(0.5 * self.rotation), 0, 0, cos(0.5 * self.rotation)],
            ]
        )

    def inv(self):
        """Apply inverse gate."""
        gate = Ryy(self.rotation)
        gate.is_inverse = not self.is_inverse 
        return gate



"""Definition of the quantum amplitude amplification Operator."""

from .operator import Operator
from qutrunk.circuit import Qureg
from qutrunk.circuit.gates import X, MCZ, All, H


class QAA(Operator):
    """Quantum Amplitude Amplification Operator.

    Args:
        iterations: The number of QAA iteration.
        marked_index: Index of the marked qubit.

    Example:
        .. code-block:: python

            from qutrunk.circuit.ops import QAA
            from qutrunk.circuit import QCircuit
            from qutrunk.circuit.gates import H, All

            circuit = QCircuit()
            qureg = circuit.allocate(4)
            All(H) * qureg
            QAA(3, 7) * qureg
            for i in range(2 ** len(qureg)):
                print(circuit.get_prob(i))

        First, the four qubits are uniformly superposed,
        and then the state value of 7 is selected as the marker value.
        Three times of QAA iterative calculation are performed.
        The result obtained after the operation is that the probability of
        the corresponding state of 7 exceeds 96%.
    """

    def __init__(self, iterations, marked_index):
        super().__init__()
        self.iterations = iterations
        self.marked_index = marked_index

    def __mul__(self, qureg: Qureg):
        """Apply the QAA operator."""
        if not isinstance(qureg, Qureg):
            raise TypeError("The operand must be Qureg.")

        if self.marked_index < 0 or self.marked_index >= 2 ** len(qureg):
            raise ValueError("The marked index value exceed 2 ** len(qureg).")

        for i in range(self.iterations):
            self._flip_process(qureg)
            self._imag_process(qureg)
            # show intermediate process
            prob_amp = qureg.circuit.get_prob(self.marked_index)
            print(f"prob of state |{self.marked_index}> = {prob_amp}")

    def _flip_process(self, qureg):
        """Flit the phase of target qubit.

        Args:
            qureg: The qureg apply flip process.
        """
        # 0b111
        bit_flip = bin(self.marked_index)
        # 00000111
        bit_flip = bit_flip[2:].zfill(len(qureg))
        # 11100000
        bit_flip = bit_flip[::-1]

        for i in range(len(bit_flip)):
            if bit_flip[i] == "0":
                X * qureg[i]

        MCZ(len(qureg) - 1) * qureg

        for i in range(len(bit_flip)):
            if bit_flip[i] == "0":
                X * qureg[i]

    def _imag_process(self, qureg):
        """Apply image process.

        Args:
            qureg: The qureg apply imag process.
        """
        All(H) * qureg
        All(X) * qureg
        MCZ(len(qureg) - 1) * qureg
        All(X) * qureg
        All(H) * qureg




"""Deserialize QuSL file object."""

from pathlib import Path

from qutrunk.circuit import QCircuit


def run_qusl_parse():
    qusl_file = Path.cwd() / "bell_pair.qusl"
    circuit = QCircuit.load(file=qusl_file)

    # run circuit
    res = circuit.run(shots=100)

    # print result
    print(res.get_measure())
    print(res.get_counts())

    return circuit


if __name__ == "__main__":
    circuit = run_qusl_parse()
    circuit.draw()



"""Run the quantum circuit using QuSprout as the backend."""


def run_check_sprout():
    from qutrunk.circuit import QCircuit
    from qutrunk.backends import BackendQuSprout
    from qutrunk.circuit.gates import H, CNOT, Measure

    # allocate
    qc = QCircuit(backend=BackendQuSprout())
    qr = qc.allocate(2)

    # apply gate
    H * qr[0]
    CNOT * (qr[0], qr[1])
    Measure * qr[0]
    Measure * qr[1]

    print("============QuSL instruction:===========")
    qc.print()
    print("===============Draw circuit=============")
    qc.draw()
    res = qc.run(shots=100)
    print("==========circuit running result=========")
    print(res.get_counts())
    print("===========circuit running info==========")
    print(res.excute_info())
    print("QuTrunk is installed successfully! You can use QuTrunk now.")


if __name__ == "__main__":
    run_check_sprout()



"""Phase Estimation Example: T-gate"""

from math import pi

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import NOT, Barrier, P, All, Measure
from qutrunk.circuit.ops import QPE


def _bin_int(itrable):
    return int("".join(map(str, reversed(itrable))), base=2)


def run_qpe(backend=None):
    """Estimate T-gate phase."""
    # allocate
    qc = QCircuit(backend=backend)
    q1, q2 = qc.allocate([3, 1])

    # Prepare our eigenstate |psi>
    NOT * q2[0]
    Barrier * q1
    Barrier * q2

    # apply QPE
    QPE(P(pi/4)) * (q1, q2)

    # measure q1
    All(Measure) * q1

    # print circuit
    # qc.print()

    # run circuit
    qc.run(shots=100)

    # print result
    print(q1.to_cl())

    # calculate the value of theta
    f = _bin_int(q1.to_cl())
    theta = f / 2 ** len(q1)
    print("θ=", theta)

    return qc


if __name__ == "__main__":
    circuit = run_qpe()






"""Hadamard gate."""

import cmath

import numpy as np

from .basicgate import BasicGate
from qutrunk.circuit import Command
from qutrunk.circuit.qubit import QuBit


class HGate(BasicGate):
    """Apply the single-qubit Hadamard gate.

    Example:
        .. code-block:: python

            H * qr[0]
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "H"

    def __or__(self, qubit):
        """Quantum logic gate operation.

        Args:
            qubit: Quantum bit applied to hadamard gate.

        Example:
            .. code-block:: python

                H * qr[0]
        """
        if not isinstance(qubit, QuBit):
            raise TypeError("The argument must be Qubit object.")

        targets = [qubit.index]
        cmd = Command(self, targets, inverse=self.is_inverse)
        self.commit(qubit.circuit, cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return 1.0 / cmath.sqrt(2.0) * np.matrix([[1, 1], [1, -1]])

    @property
    def label(self):
        """A label for identifying the gate."""
        self.__str__()

    def inv(self):
        """Apply inverse gate."""
        gate = HGate()
        gate.is_inverse = not self.is_inverse 
        return gate

    def ctrl(self, ctrl_cnt=1):
        """Apply controlled gate.
        
        Args:
            ctrl_cnt: The number of control qubits, default: 1.
        """
        if ctrl_cnt > 1:
            raise ValueError("H gate do not support multiple control bits.")
        gate = CHGate()
        gate.is_inverse = self.is_inverse
        return gate


H = HGate()


class CHGate(BasicGate):
    """Controlled-Hadamard gate.

    Example:
        .. code-block:: python

            CH * (qr[0], qr[1])
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "CH"

    def __or__(self, qubits):
        """Quantum logic gate operation.

        Args:
            qubits: qubits[0] is control qubit, qubits[1] is target qubit.

        Example:
            .. code-block:: python

                CH * (qr[0], qr[1])

        Raises:
            AttributeError: If the argument is not a Qubit object.
        """
        if not all(isinstance(qubit, QuBit) for qubit in qubits):
            # TODO:need to improve.
            raise NotImplementedError("The argument must be Qubit object.")

        if len(qubits) != 2:
            # TODO:need to improve.
            raise AttributeError("Parameter error: One controlled and one target qubit are required.")

        self.qubits = qubits
        controls = [qubits[0].index]
        targets = [qubits[1].index]
        cmd = Command(self, targets, controls, inverse=self.is_inverse)
        self.commit(qubits[0].circuit, cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement.

        Args:
            qubits: qubits[0] is control qubit, qubits[1] is target qubit.
        """
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        _sqrt2o2 = 1 / cmath.sqrt(2)
        return np.array(
            [
                [_sqrt2o2, 0, _sqrt2o2, 0],
                [0, 1, 0, 0],
                [_sqrt2o2, 0, -_sqrt2o2, 0],
                [0, 0, 0, 1],
            ]
        )

    def inv(self):
        """Apply inverse gate."""
        gate = CHGate()
        gate.is_inverse = not self.is_inverse 
        return gate


CH = CHGate()



import math

import numpy as np

from .basicgate import BasicRotateGate
from qutrunk.circuit import Command
from qutrunk.circuit.qubit import QuBit


class R(BasicRotateGate):
    """Rotation alpha around the cos(beta)x + sin(beta)y axis.

    Args:
        theta: The angle to rotate.
        phi: Define the axis cos(beta)x + sin(beta)y.

    Example:
        .. code-block:: python

            R(theta, phi) * qr[0]
    """

    def __init__(self, theta, phi):
        if theta is None or phi is None:
            raise ValueError("The argument cannot be empty.")
        super().__init__()
        self.theta = theta
        self.phi = phi

    def __str__(self):
        return "R"

    def __or__(self, qubit):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to apply R gate.

        Example:
            .. code-block:: python

                R(theta, phi) * qr[0]

        Raises:
            TypeError: If the argument is not a Qubit object.
        """
        if not isinstance(qubit, QuBit):
            raise TypeError("The argument must be Qubit object.")

        targets = [qubit.index]
        cmd = Command(
            self, targets, rotation=[self.theta, self.phi], inverse=self.is_inverse
        )
        self.commit(qubit.circuit, cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        # TODO:define have problem.
        theta, phi = float(self.theta), float(self.phi)
        cos = math.cos(theta / 2)
        sin = math.sin(theta / 2)
        exp_m = np.exp(-1j * phi)
        exp_p = np.exp(1j * phi)
        return np.array([[cos, -1j * exp_m * sin], [-1j * exp_p * sin, cos]])

    def inv(self):
        """Apply inverse gate."""
        gate = R(self.theta, self.phi)
        gate.is_inverse = not self.is_inverse 
        return gate



"""Self-decrement operation example."""

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import Measure, All
from qutrunk.circuit.ops import Classical
from qutrunk.circuit.ops import DEC


def decrement(num_qubits, init_value):
    circuit = QCircuit()
    qr = circuit.allocate(num_qubits)

    Classical(init_value) * qr

    DEC * qr

    All(Measure) * qr
    res = circuit.run()
    print(res.get_outcome())

    return circuit


if __name__ == "__main__":
    # Run locally
    circuit = decrement(4, 1)

    # Draw quantum circuit
    circuit.draw()




import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicGate, Observable, PauliType
from qutrunk.circuit.qubit import QuBit


class XGate(BasicGate, Observable):
    """Apply the single-qubit Pauli-X (also known as the X, sigma-X, NOT or bit-flip) gate.

    Example:
        .. code-block:: python

            X * qr[0]
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "X"

    def __or__(self, qubit):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to apply X gate.

        Example:
            .. code-block:: python

                X * qr[0]

        Raises:
            NotImplementedError: If the argument is not a Qubit object.
        """
        if not isinstance(qubit, QuBit):
            raise TypeError("The argument must be Qubit object.")

        targets = [qubit.index]
        cmd = Command(self, targets, inverse=self.is_inverse)
        self.commit(qubit.circuit, cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix([[0, 1], [1, 0]])

    def __call__(self, target):
        """
        Get Observable data.

        Args:
            target: The observed qubit.

        Returns:
            The observed data list, each item contains op type and target qubit, \
                e.g: [{"oper_type": 1, "target": 0}].
        """
        pauli = {}
        pauli["oper_type"] = PauliType.PAULI_X.value
        pauli["target"] = target.index
        return pauli

    def inv(self):
        """Apply inverse gate."""
        gate = XGate()
        gate.is_inverse = not self.is_inverse
        return gate

    def ctrl(self, ctrl_cnt=1):
        """Apply controlled gate.
        
        Args:
            ctrl_cnt: The number of control qubits, default: 1.
        """
        gate = MCX(ctrl_cnt)
        gate.is_inverse = self.is_inverse
        return gate


PauliX = X = NOT = XGate()


class MCX(BasicGate):
    """Multi-control X(NOT) gate.

    Args:
        ctrl_cnt: The number of control qubits.

    Example:
        .. code-block:: python

            MCX(2) * (qr[0], qr[1], qr[2]) # qr[0], qr[1] are control qubits, qr[2] is target qubit
            MCX(3) * (qr[0], qr[1], qr[2], qr[3])
    """

    def __init__(self, ctrl_cnt=1):
        """Create new MCX gate."""
        super().__init__()
        self.ctrl_cnt = ctrl_cnt

    def __str__(self):
        return "MCX"

    def __or__(self, qubits):
        """Quantum logic gate operation.

        Args:
            qubits: The left self.ctrl_cnt qubits are control qubits, the rest right bits are target qubits.

        Example:
            .. code-block:: python

                MCX(2) * (qr[0], qr[1], qr[2]) # qr[0], qr[1] are control qubits, qr[2] is target qubit
                MCX(3) * (qr[0], qr[1], qr[2], qr[3])
        """
        if not all(isinstance(qubit, QuBit) for qubit in qubits):
            raise TypeError("The argument must be Qubit object.")

        if len(qubits) <= self.ctrl_cnt:
            raise ValueError("The parameter miss controlled or target qubit(s).")

        self.qubits = qubits
        controls = [q.index for q in qubits[0 : self.ctrl_cnt]]
        targets = [q.index for q in qubits[self.ctrl_cnt :]]
        cmd = Command(self, targets, controls, inverse=self.is_inverse)
        self.commit(qubits[0].circuit, cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        if self.ctrl_cnt == 1:
            return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])

    def inv(self):
        """Apply inverse gate."""
        gate = MCX(self.ctrl_cnt)
        gate.is_inverse = not self.is_inverse
        return gate


CX = CNOT = MCX(1)
Toffoli = MCX(2)



"""Resources required for the statistics of quantum circuits:
number of quantum bits, number of quantum gates; Quantum circuits are time-consuming to compile and run.
"""
import time


class Counter:
    """Resources of circuit.

    Args:
        circuit: A quantum circuit to statistics the used resource.
    """

    def __init__(self, circuit):
        self.circuit = circuit
        self.qubits = 0
        self.quantum_gates = 0
        self.start_time = time.time()
        self.total_time = 0
        self.backend_time = 0
        self.qutrunk_time = 0

    def acc_run_time(self, elapsed):
        """Accumulate backend running time.

        Args:
            elapsed: The elapsed time.
        """
        self.backend_time += elapsed

    def finish(self):
        """Statistics time and gates when circuit running finish."""
        self.total_time = time.time() - self.start_time
        self.qutrunk_time = self.total_time - self.backend_time
        self.quantum_gates = self.circuit.num_gates

    def __repr__(self):
        return f"Counter(quit={self.qubits})"

    def show_verbose(self):
        """Print the resource info of circuit."""
        print("==================Counter==================")
        print(self)
        print("qubits =", self.qubits)
        print("quantum_gates =", self.quantum_gates)
        print("total_time =", self.total_time)
        print("qutrunk_time =", self.qutrunk_time)
        print("backend_time =", self.backend_time)



"""Deserialize OpenQASM file object."""

import os
from pathlib import Path

from qutrunk.circuit import QCircuit


def run_openqasm_parse():
    cur_path = os.path.dirname(os.path.abspath(__file__))
    qasm_file = Path(cur_path) / "bell_pair.qasm"
    circuit = QCircuit.load(file=qasm_file, format="openqasm")

    # run circuit
    res = circuit.run(shots=100)

    # print result
    print(res.get_measure())
    print(res.get_counts())

    return circuit


if __name__ == "__main__":
    circuit = run_openqasm_parse()
    circuit.draw()



"""Quantum circuit convert to DAG."""
from qutrunk.dagcircuit.dagcircuit import DAGCircuit
from qutrunk.circuit import QuBit, CBit


def circuit_to_dag(circuit):
    """Convert circuit to DAGCircuit.

     Args:
        circuit: The input circuit.

    Return:
        DAGCircuit: The DAG representing the input circuit.
    """
    dagcircuit = DAGCircuit()
    dagcircuit.name = circuit.name

    dagcircuit.add_qubits(circuit.qubits)
    dagcircuit.add_cbits(circuit.cbits)

    for cmd in circuit.cmds:
        qargs, cargs = __operation_command(cmd, circuit)
        dagcircuit.apply_operation_back(cmd, qargs, cargs)

    return dagcircuit


def __operation_command(command, circuit):
    """Rebuild instruction."""
    qargs = []
    cargs = []

    for c in command.controls:
        qubit = QuBit(circuit.qreg, index=c)
        qargs.append(qubit)

    for c in command.targets:
        qubit = QuBit(circuit.qreg, index=c)
        qargs.append(qubit)

    if str(command.gate) == "Measure":
        cbit = CBit(circuit.creg, index=command.targets[0])
        cargs.append(cbit)

    return qargs, cargs



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qutrunk',
 'qutrunk.backends',
 'qutrunk.backends.braket',
 'qutrunk.backends.ibm',
 'qutrunk.backends.qusprout',
 'qutrunk.circuit',
 'qutrunk.circuit.gates',
 'qutrunk.circuit.ops',
 'qutrunk.converters',
 'qutrunk.dagcircuit',
 'qutrunk.example',
 'qutrunk.qasm',
 'qutrunk.qasm.node',
 'qutrunk.sim',
 'qutrunk.sim.local',
 'qutrunk.sim.qusprout',
 'qutrunk.sim.qusprout.code',
 'qutrunk.sim.qusprout.qusprout',
 'qutrunk.sim.qusprout.qusproutdata',
 'qutrunk.test',
 'qutrunk.test.backend',
 'qutrunk.test.circuit',
 'qutrunk.test.gate',
 'qutrunk.test.tools',
 'qutrunk.tools',
 'qutrunk.visualizations']

package_data = \
{'': ['*'],
 'qutrunk': ['config/*'],
 'qutrunk.example': ['jupyter/*'],
 'qutrunk.qasm': ['libs/*'],
 'qutrunk.sim.qusprout': ['idl/*']}

install_requires = \
['amazon-braket-sdk>=1.32.0,<2.0.0',
 'networkx>=2.8,<3.0',
 'numpy>=1.21.0,<2.0.0',
 'ply>=3.11,<4.0',
 'pyyaml>=6.0,<7.0',
 'requests>=2.27.1,<3.0.0',
 'retworkx>=0.11.0,<0.12.0',
 'thrift>=0.15.0,<0.16.0']

setup_kwargs = {
    'name': 'qutrunk',
    'version': '0.1.12',
    'description': 'qutrunk is an open source library for quantum computing.',
    'long_description': '## QuTrunk\n\n[![Documentation Status](https://img.shields.io/badge/docs-latest-brightgreen.svg)](http://developer.queco.cn/qutrunk_api/)\n[![License](https://img.shields.io/badge/license-Apache%202-blue.svg)](LICENSE)\n[![Download Code](https://img.shields.io/badge/download-zip-green.svg)](https://github.com/queco-quantum/qutrunk/archive/refs/heads/main.zip)\n\n\n### **概述**\n---\n* QuTrunk 是启科量子自主研发的一款免费、开源、跨平台的量子计算编程框架，包括量子编程API、量子命令转译、量子计算后端接口等。\n* QuTrunk 使用 Python 作为宿主语言，利用 Python 的语法特性实现针对量子程序的 DSL (领域专用语言)，所有使用 Python 编程的 IDE 均可使用安装。\n* QuTrunk 基于量子逻辑门、量子线路等概念提供量子编程所需各类 API，这些 API 由相应的模块实现。例如 QCircuit 实现量子线路，Qubit 实现量子比特，Qureg 实现量子寄存器，Command 实现每个量子门操作的指令，Backend 实现运行量子线路的后端模块，gate 模块实现各类基础量子门操作。\n* QuTrunk 还可以作为其他上层量子计算应用的基础，例如：量子算法、量子可视化编程、量子机器学习等。\n\nQuTrunk内部模块划分及层次结构如下：  \n\n<div align=center>\n<img src="./resource/qutrunk.png"/>\n</div>\n\n\n### **核心模块**\n---\n* cicuit: 量子线路，通过应用各类基础门操作以及算符操作构建量子线路，代表了整个量子算法的实现。\n* qubit: 代表单个量子比特，是量子门和量子算符操作的目标对象。\n* qureg: 用于申请量子计算资源，维护若干个量子比特，用于实现某个具体的量子算法。\n* gate: 量子逻辑门模块，提供各类基础量子门操作，包括:*H*, *X*, *Y*, *Z*，*P*, *R*, *Rx*, *Ry*, *Rz*, *S*, *Sdg*, *T*, *Tdg*, *CNOT*, *Toffoli*, *Swap*等。\n* operator: 量子算符操作，通过若干基础量子门实现某些通用量子操作，比如振幅放大QAA, 量子傅立叶变换QFT等。\n* command: 对量子线路中所有门级操作做参数化处理，对接目标后端模块，用于运行整个量子线路。\n* qasm: 兼容OpenQASM 2.0标准，实现量子线路到OpenQASM指令的序列化和反序列化。\n* qusl: QuTrunk量子汇编标准，实现与qasm类似功能。\n* backend: 量子计算后端模块，用于执行量子线路，支持Python本地后端，qusprout和qusaas两种远程后端以及第三方后端(目前支持IBM)。\n* qusprout: 对接启科研制的qubox设备，使用经典计算资源并针对量子计算特点做优化，提供高性能量子模拟计算服务。\n* qusaas: 对接启科量子计算云平台，接入多种量子计算资源，包括经典计算资源，离子阱量子计算机（研发中）。\n\n\n### 主要特点\n---\n* 基于量子逻辑门、量子算符和量子线路实现量子程序开发。\n* 提供QuSL量子汇编指令标准，QuSL量子汇编与Python代码完全兼容。\n* 设备独立，同一个量子线路只需替换后端类型即可以在不同的量子后端上运行。\n* 提供多种量子计算体验，本地量子计算提供Python计算后端，远程后端提供OMP多线程、MPI多节点并行、GPU加速等计算模式，同时预留了接口对接启科量子自行研制的离子阱量子计算机。\n* 兼容多种量子汇编指令格式：OpenQASM 2.0标准和QuSL汇编标准。\n* 支持量子可视化编程（需要配合启科量子研发的量子集成开发环境 QuBranch）。\n\n\n### **下载和安装**\n---\n#### **pip安装** \n\nQuTrunk 已发布于 PyPI 官网，可以通过 pip 命令进行安装。\n注意在正式使用 QuTurnk 之前，您需要先安装 Python（版本 3.8+）。\n\n  ```python\n  pip install qutrunk\n  ```\n\n验证QuTrunk是否安装成功，打开终端进入python交互模式，执行如下语句：\n\n``` python\nimport qutrunk\nqutrunk.run_check()\n```\n输出结果为："QuTrunk is installed successfully! You can use QuTrunk now."表明QuTrunk安装成功。\n\n\n### **示例代码**\n---\n以下示例展示了利用 QuTrunk 运行 bell-pair 量子算法：\n\n  ```python\n  # import package\n  from qutrunk.circuit import QCircuit\n  from qutrunk.circuit.gates import H, CNOT, Measure, All\n\n  # allocate resource\n  qc = QCircuit()\n  qr = qc.allocate(2) \n\n  # apply quantum gates\n  H * qr[0]   \n  CNOT * (qr[0], qr[1])\n  All(Measure) * qr\n\n  # print circuit\n  qc.print()   \n  # run circuit\n  res = qc.run(shots=1024) \n  # print result\n  print(res.get_counts()) \n  # draw circuit\n  qc.draw()\n  ```\n\n运行结果：\n<div>\n<img src="./resource/bell_pair.png"/>\n</div>\n\n### **量子可视化编程**  \n---\nQuBranch是由启科量子基于vscode开发的量子编程集成开发环境, QuTrunk与QuBranch相互配合可以实现量子可视化编程,\n具体步骤参见[量子可视化编程](http://developer.queco.cn/learn/doc/detail?id=12&childrenid=14)\n\n### **文档**\n---\n* [QuTrunk 快速上手教程](http://developer.queco.cn/learn/doc/detail?id=12&childrenid=14)\n* [QuTrunk API](http://developer.queco.cn/qutrunk_api/)\n\n\n### **如何参与开发**\n---\n1. 阅读源代码，了解我们当前的开发方向\n2. 找到自己感兴趣的功能或模块\n3. 进行开发，开发完成后自测功能是否正确\n4. Fork代码库，将修复代码提交到fork的代码库\n5. 发起pull request\n6. 更多详情请参见[链接](./CONTRIBUTING.md)\n\n\n### **许可证**\n---\nQuTrunk是自由和开源的，在Apache 2.0许可证版本下发布。\n',
    'author': 'qudoorzh2022',
    'author_email': 'qudoorzh2022@163.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://www.qudoor.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import numpy as np

from .basicgate import BasicGate
from qutrunk.circuit import Command
from qutrunk.circuit.qubit import QuBit


class iSwapGate(BasicGate):
    """Performs a iSwap gate between qubit1 and qubit2.

    Example:
        .. code-block:: python

            iSwap * (qr[0], qr[1])
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "iSwap"

    def __or__(self, qubits):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to aplly iSwap gate.

        Example:
            .. code-block:: python

                iSwap * (qr[0], qr[1])

        Raises:
            NotImplementedError: If the argument is not a Qubit object.
            AttributeError: If the qubits should not be two.
        """
        if not all(isinstance(qubit, QuBit) for qubit in qubits):
            raise TypeError("The argument must be Qubit object.")

        if len(qubits) != 2:
            raise ValueError("Parameter Error: Two target bits are required.")

        self.qubits = qubits
        targets = [q.index for q in qubits]
        cmd = Command(self, targets, inverse=self.is_inverse)
        self.commit(qubits[0].circuit, cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix([[1, 0, 0, 0], [0, 0, 1j, 0], [0, 1j, 0, 0], [0, 0, 0, 1]])

    def inv(self):
        """Apply inverse gate."""
        gate = iSwapGate()
        gate.is_inverse = not self.is_inverse
        return gate


iSwap = iSwapGate()



"""Testing the functionality of exporting quantum circuits in OpenQASM format."""
from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, Measure, CNOT


def test_print_qasm():
    """Test the print_qasm function in QCircuit."""
    qc = QCircuit()
    qreg = qc.allocate(2)

    H * qreg[0]
    CNOT * (qreg[0], qreg[1])
    Measure * qreg[0]
    Measure * qreg[1]

    qc.print_qasm(file="b.qasm")
    expect_out = ['OPENQASM 2.0;\n', 'include "qulib1.inc";\n', 'qreg q[2];\n', 'creg c[2];\n', 'h q[0];\n', 'cx q[0],q[1];\n', 'measure q[0] -> c[0];\n', 'measure q[1] -> c[1];\n']

    with open("b.qasm", "r") as stream:
        circuit_out = stream.readlines()

    assert circuit_out == expect_out



"""Testing the functionality of exporting quantum circuits in Qusl format."""
from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, Measure, CNOT


def test_print_qusl():
    """Test the print function in QCircuit."""
    qc = QCircuit()
    qreg = qc.allocate(2)

    H * qreg[0]
    CNOT * (qreg[0], qreg[1])
    Measure * qreg[0]
    Measure * qreg[1]

    qc.print(file="a.qusl")
    expect_out = '"code": ["H * q[0]\\n", "MCX * (q[0], q[1])\\n", "Measure * q[0]\\n", "Measure * q[1]\\n"]'

    with open("a.qusl", "r") as stream:
        container = stream.readline()
        circuit_out = container[container.rfind("code")-1:-1]

    assert circuit_out == expect_out



"""U1 Gate."""

import numpy as np

from qutrunk.circuit import Command
from qutrunk.circuit.qubit import QuBit
from .basicgate import BasicGate, BasicRotateGate


class U1(BasicGate):
    """U1 gate.

    Args:
        alpha: Rotation angle.

    Example:
        .. code-block:: python

            U1(pi/2) * qr[0]
    """

    def __init__(self, alpha):
        """
        Args:
            alpha: Rotation angle.
        """
        if alpha is None:
            raise NotImplementedError("The argument cannot be empty.")
        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "U1"

    def __or__(self, qubit):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to apply U1 gate.

        Example:
            .. code-block:: python

                U1(pi/2) * qr[0]

        Raises:
            TypeError: If the argument is not a Qubit object.
        """
        if not isinstance(qubit, QuBit):
            raise TypeError("The argument must be Qubit object.")

        targets = [qubit.index]
        cmd = Command(self, targets, rotation=[self.rotation], inverse=self.is_inverse)
        self.commit(qubit.circuit, cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, \
            reuse __or__ operator implement."""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        lam = float(self.rotation)
        return np.array([[1, 0], [0, np.exp(1j * lam)]])

    def inv(self):
        """Apply inverse gate."""
        gate = U1(self.rotation)
        gate.is_inverse = not self.is_inverse
        return gate

    def ctrl(self, ctrl_cnt=1):
        """Apply controlled gate.
        
        Args:
            ctrl_cnt: The number of control qubits, default: 1.
        """
        if ctrl_cnt > 1:
            raise ValueError("U1 gate do not support multiple control bits.")
        gate = CU1(self.rotation)
        gate.is_inverse = self.is_inverse
        return gate


class CU1(BasicRotateGate):
    """Control U1 gate.

    Args:
        alpha: Rotation angle.

    Example:
        .. code-block:: python

           CU1(pi/2) * (qr[0], qr[1])
    """

    def __init__(self, alpha):
        """Create new CU1 gate.
        Args:
            alpha: Rotation angle.
        """
        if alpha is None:
            raise TypeError("The argument cannot be empty.")
        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "CU1"

    def __or__(self, qubits):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to apply CU1 gate.

        Example:
            .. code-block:: python

                CU1(pi/2) * (qr[0], qr[1])

        Raises:
            TypeError: If the argument is not a Qubit object.
            AttributeError: If the qubits should not be two.
        """
        if not all(isinstance(qubit, QuBit) for qubit in qubits):
            raise TypeError("The argument must be Qubit object.")

        if len(qubits) != 2:
            raise ValueError("Parameter error: One controlled and one target qubit are required.")

        controls = [qubits[0].index]
        targets = [qubits[1].index]
        cmd = Command(
            self, targets, controls, inverse=self.is_inverse, rotation=[self.rotation]
        )
        self.commit(qubits[0].circuit, cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        # TODO: definition have problem.
        half_alpha = float(self.rotation)
        return np.matrix(
            [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, np.exp(1j * half_alpha)]]
        )

    def inv(self):
        """Apply inverse gate."""
        gate = CU1(self.rotation)
        gate.is_inverse = not self.is_inverse
        return gate



"""Self-decrement operation."""

from qutrunk.circuit import Qureg
from qutrunk.circuit.ops.operator import Operator
from qutrunk.circuit.gates import X, MCX


class Decrement(Operator):
    """Self-decrement operation.

    Example:
        .. code-block:: python

            from qutrunk.circuit import QCircuit
            from qutrunk.circuit.gates import Measure, All
            from qutrunk.circuit.ops import Classical
            from qutrunk.circuit.ops import DEC

            circuit = QCircuit()
            qr = circuit.allocate(4)
            Classical(4) * qr
            DEC * qr
            All(Measure) * qr
            res = circuit.run()
            print(res.get_outcome())

    """

    def __init__(self):
        super().__init__()

    def __mul__(self, qr: Qureg):
        if not isinstance(qr, Qureg):
            raise TypeError("The operand must be Qureg.")

        num_qubits = len(qr)

        X * qr[0]
        ctrl = []
        for i in range(1, num_qubits):
            for j in range(i):
                ctrl.append(qr[j])
            MCX(i) * (*ctrl, qr[i])
            ctrl = []


DEC = Decrement()



"""Helper function for converting a dag to a circuit."""

from qutrunk.circuit import QCircuit
from qutrunk.dagcircuit.dagnode import DAGOpNode


def dag_to_circuit(dag):
    """Build a ``QCircuit`` object from a ``DAGCircuit``.

    Args:
        dag (DAGCircuit): The input dag.

    Return:
        QCircuit: The circuit representing the input dag.
    """

    circuit = QCircuit()
    if dag.name:
        circuit.name = dag.name
    circuit.allocate(len(dag.qubits))
    circuit.metadata = dag.metadata

    for node in dag.topological_op_nodes():
        if hasattr(node.op, "definition"):
            unroll_custom_gate(circuit, node.op.definition, node.qargs, node.cargs)
        else:
            append_node(circuit, node)

    return circuit


def append_node(circuit: QCircuit, node: DAGOpNode):
    """
    Interpret node and apply to circuit
    support both single qubit and multi qubit gate
    as measure gate and all gate only take qargs, cargs is omitted

    Args:
        circuit: Generated circuit.
        node: Dag node tobe interpreted.
    """
    gate = node.op
    qubit = node.qargs
    for qb in qubit:
        qb.circuit = circuit
    if len(qubit) == 1:
        qubit = qubit[0]
    gate * qubit


def unroll_custom_gate(circuit, nodes, qargs, cargs):
    """Unroll custom gate in QASM file, nested custom gate is supported.

    Args:
        circuit (QCircuit): Circuit to append gate.
        nodes: Operation node from dag.
        qargs: Qargs for the gate.
        cargs: Cargs for the gate.
    """
    for inst in nodes:
        op = inst[0]
        inner_qargs = inst[1]
        inner_cargs = inst[2]

        real_qargs = []
        for qb in inner_qargs:
            real_qargs.append(qargs[qb.index])

        real_cargs = []
        for cb in inner_cargs:
            real_cargs.append(cargs[cb.index])

        if hasattr(op, "definition"):
            unroll_custom_gate(circuit, op.definition, real_qargs, real_cargs)
        else:
            append_node(circuit, DAGOpNode(op, real_qargs, real_cargs))



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""AST (abstract syntax tree) to DAG (directed acyclic graph) converter.

Acts as an OpenQASM interpreter.
"""
import re
from collections import OrderedDict

from qutrunk.circuit import CReg, Qureg
from qutrunk.circuit.gates import BasicGate, MeasureGate, BarrierGate, MCX
from qutrunk.converters.mapping import qutrunk_standard_gate
from qutrunk.dagcircuit import DAGCircuit
from qutrunk.exceptions import QuTrunkError
from qutrunk.qasm.node.real import Real


def ast_to_dag(ast):
    """Build a ``DAGCircuit`` object from an AST ``Node`` object.

    Args:
        Ast (Program): A Program Node of an AST (parser's output).

    Return:
        DAGCircuit: The DAG representing an OpenQASM's AST.

    Raises:
        QuTrunkError: If the AST is malformed.
    """
    dag = DAGCircuit()
    AstInterpreter(dag)._process_node(ast)

    return dag


class AstInterpreter:
    """Interprets an OpenQASM by expanding subroutines and unrolling loops.

    Args:
        dag: The DAG representing an OpenQASM's AST.
    """

    def __init__(self, dag):
        """Initialize interpreter's data."""
        # DAG object to populate
        self.dag = dag
        # OPENQASM version number (ignored for now)
        self.version = 0.0
        # Dict of gates names and properties
        self.gates = OrderedDict()
        # Keeping track of conditional gates
        self.condition = None
        # List of dictionaries mapping local parameter ids to expression Nodes
        self.arg_stack = [{}]
        # List of dictionaries mapping local bit ids to global ids (name, idx)
        self.bit_stack = [{}]

    def _process_bit_id(self, node):
        """Process an Id or IndexedId node as a bit or register type.

        Returns:
            Return a list of tuples (Register,index).

        Raises:
            QuTrunkError: If the AST is malformed.
        """
        reg = None

        if node.name in self.dag.qregs:
            reg = self.dag.qregs[node.name]
        elif node.name in self.dag.cregs:
            reg = self.dag.cregs[node.name]
        else:
            raise QuTrunkError(
                "expected qreg or creg name:",
                "line=%s" % node.line,
                "file=%s" % node.file,
            )

        if node.type == "indexed_id":
            # An indexed bit or qubit
            return [reg[node.index]]
        elif node.type == "id":
            # A qubit or qreg or creg
            if not self.bit_stack[-1]:
                # Global scope
                return list(reg)
            else:
                # local scope
                if node.name in self.bit_stack[-1]:
                    return [self.bit_stack[-1][node.name]]
                raise QuTrunkError(
                    "expected local bit name:",
                    "line=%s" % node.line,
                    "file=%s" % node.file,
                )
        return None

    def _process_custom_unitary(self, node):
        """Process a custom unitary node.

        Raises:
            QuTrunkError: If the AST is malformed.
        """
        name = node.name
        if node.arguments is not None:
            args = self._process_node(node.arguments)
        else:
            args = []
        bits = [
            self._process_bit_id(node_element) for node_element in node.bitlist.children
        ]

        if name in self.gates:
            self._arguments(name, bits, args)
        else:
            raise QuTrunkError(
                "internal error undefined gate:",
                f"line={node.line}",
                f"file={node.file}",
            )

    def _process_u(self, node):
        """Process a U gate node."""
        args = self._process_node(node.arguments)
        bits = [self._process_bit_id(node.bitlist)]

        self._arguments("u", bits, args)

    def _arguments(self, name, bits, args):
        """Gate arguments."""
        gargs = self.gates[name]["args"]
        gbits = self.gates[name]["bits"]

        maxidx = max(map(len, bits))
        for idx in range(maxidx):
            self.arg_stack.append({gargs[j]: args[j] for j in range(len(gargs))})
            # Only index into register arguments.
            element = [idx * x for x in [len(bits[j]) > 1 for j in range(len(bits))]]
            self.bit_stack.append(
                {gbits[j]: bits[j][element[j]] for j in range(len(gbits))}
            )
            self._create_dag_op(
                name,
                [self.arg_stack[-1][s].sym() for s in gargs],
                [self.bit_stack[-1][s] for s in gbits],
            )
            self.arg_stack.pop()
            self.bit_stack.pop()

    def _process_gate(self, node, opaque=False):
        """Process a gate node.

        If opaque is True, process the node as an opaque gate node.
        """
        self.gates[node.name] = {}
        de_gate = self.gates[node.name]
        de_gate["print"] = True  # default
        de_gate["opaque"] = opaque
        de_gate["n_args"] = node.n_args()
        de_gate["n_bits"] = node.n_bits()
        if node.n_args() > 0:
            de_gate["args"] = [element.name for element in node.arguments.children]
        else:
            de_gate["args"] = []
        de_gate["bits"] = [c.name for c in node.bitlist.children]
        if node.name in qutrunk_standard_gate:
            return
        if opaque:
            de_gate["body"] = None
        else:
            de_gate["body"] = node.body

    def _process_cnot(self, node):
        """Process a CNOT gate node.

        Raises:
            QuTrunkError: If the AST is malformed.
        """
        id0 = self._process_bit_id(node.children[0])
        id1 = self._process_bit_id(node.children[1])
        if not (len(id0) == len(id1) or len(id0) == 1 or len(id1) == 1):
            raise QuTrunkError(
                "internal error: qreg size mismatch:",
                f"line={node.line}",
                f"file={node.file}",
            )

        maxidx = max([len(id0), len(id1)])
        for idx in range(maxidx):
            cx_gate = MCX()
            if len(id0) > 1 and len(id1) > 1:
                self.dag.apply_operation_back(cx_gate, [id0[idx], id1[idx]], [])
            elif len(id0) > 1:
                self.dag.apply_operation_back(cx_gate, [id0[idx], id1[0]], [])
            else:
                self.dag.apply_operation_back(cx_gate, [id0[0], id1[idx]], [])

    def _process_measure(self, node):
        """Process a measurement node."""
        id0 = self._process_bit_id(node.children[0])
        id1 = self._process_bit_id(node.children[1])
        if len(id0) != len(id1):
            raise QuTrunkError(
                "internal error: reg size mismatch:",
                f"line={node.line}",
                f"file={node.file}",
            )

        for idx, idy in zip(id0, id1):
            meas_gate = MeasureGate()
            meas_gate.name = "measure"
            self.dag.apply_operation_back(meas_gate, [idx], [idy])

    # note: qutrunk not support if condition
    def _process_if(self, node):
        """Process an if node."""
        pass

    def _process_children(self, node):
        """Call process_node for all children of node."""
        for kid in node.children:
            self._process_node(kid)

    def _process_node(self, node):
        """Carry out the action associated with a node.

        Raises:
            QuTrunkError: If the AST is malformed.
        """
        if node.type == "program":
            self._process_children(node)

        elif node.type == "qreg":
            qreg = Qureg(size=node.index, name=node.name)
            self.dag.add_qreg(qreg)

        elif node.type == "creg":
            creg = CReg(size=node.index, name=node.name)
            self.dag.add_creg(creg)

        elif node.type == "id":
            raise QuTrunkError("internal error: _process_node on id")

        elif node.type == "int":
            raise QuTrunkError("internal error: _process_node on int")

        elif node.type == "real":
            raise QuTrunkError("internal error: _process_node on real")

        elif node.type == "indexed_id":
            raise QuTrunkError("internal error: _process_node on indexed_id")

        elif node.type == "id_list":
            # We process id_list nodes when they are leaves of barriers.
            return [
                self._process_bit_id(node_children) for node_children in node.children
            ]

        elif node.type == "primary_list":
            # We should only be called for a barrier.
            return [self._process_bit_id(m) for m in node.children]

        elif node.type == "gate":
            self._process_gate(node)

        elif node.type == "custom_unitary":
            self._process_custom_unitary(node)

        elif node.type == "universal_unitary":
            self._process_u(node)

        elif node.type == "cnot":
            self._process_cnot(node)

        elif node.type == "expression_list":
            return node.children

        elif node.type == "binop":
            raise QuTrunkError("internal error: _process_node on binop")

        elif node.type == "prefix":
            raise QuTrunkError("internal error: _process_node on prefix")

        elif node.type == "measure":
            self._process_measure(node)

        elif node.type == "format":
            self.version = node.version()

        elif node.type == "barrier":
            ids = self._process_node(node.children[0])
            qubits = []
            for qubit in ids:
                for j, _ in enumerate(qubit):
                    qubits.append(qubit[j])
            self.dag.apply_operation_back(BarrierGate(), qubits, [])

        elif node.type == "reset":
            id0 = self._process_bit_id(node.children[0])
            for i, _ in enumerate(id0):
                # qutrunk not support reset yet
                pass

        elif node.type == "if":
            self._process_if(node)

        elif node.type == "opaque":
            self._process_gate(node, opaque=True)

        elif node.type == "external":
            raise QuTrunkError("internal error: _process_node on external")

        else:
            raise QuTrunkError(
                "internal error: undefined node type:",
                f"line={node.line}",
                f"file={node.file}",
            )

        return None

    def _gate_rules_to_qutrunk_circuit(self, node, params):
        """From a gate definition in qasm, to a QCircuit format."""
        rules = []
        qreg = Qureg(size=node["n_bits"])
        bit_args = {node["bits"][i]: q for i, q in enumerate(qreg)}
        exp_args = {node["args"][i]: Real(q) for i, q in enumerate(params)}

        for child_op in node["body"].children:
            qparams = []
            eparams = []
            for param_list in child_op.children[1:]:
                if param_list.type == "id_list":
                    qparams = [bit_args[param.name] for param in param_list.children]
                elif param_list.type == "expression_list":
                    for param in param_list.children:
                        eparams.append(param.sym(nested_scope=[exp_args]))
            op = self._create_op(child_op.name, params=eparams)
            rules.append((op, qparams, []))

        sub_circ = []
        for instr, qargs, cargs in rules:
            sub_circ.append((instr, qargs, cargs))
        return sub_circ

    def _create_dag_op(self, name, params, qargs):
        """Create a DAG node out of a parsed AST op node.

        Args:
            name (str): Operation name to apply to the DAG.
            params (list): Op parameters.
            qargs (list(Qubit)): Qubits to attach to.

        Raises:
            QuTrunkError: If encountering a non-basis opaque gate.
        """
        op = self._create_op(name, params)
        self.dag.apply_operation_back(op, qargs, [])

    def _create_op(self, name, params):
        # check whether gate name indicate multi control bit
        name, ctr_cnt = count_ctrl(name)
        if ctr_cnt:
            params = [ctr_cnt] + params

        if name in qutrunk_standard_gate:
            op = qutrunk_standard_gate[name](*params)
            op.name = name
        elif name in self.gates:
            op = BasicGate(*params)
            op.name = name
            if not self.gates[name]["opaque"]:
                # call a custom gate (otherwise, opaque)
                op.definition = self._gate_rules_to_qutrunk_circuit(
                    self.gates[name], params=params
                )
        else:
            raise QuTrunkError(f"unknown operation for ast node name {name}.")
        return op


def count_ctrl(name):
    """
    Check whether gate name indicate multi control bit
    if yes, replace cc/c[0-9]+ with 'mc' and get control bits count
    control bits always at the beginning in qargs.

    Args:
        name (str): Origin operation name in OpenQASM.

    Return:
        name (str): Operation name after process.
        cnt (int): Control bits count.
    """
    cnt = 0
    multi_ctrl_reg = r"c(c|\d+)"
    match = re.match(multi_ctrl_reg, name)
    if match:
        cnt_str = match.group(1)
        if cnt_str == "c":
            cnt = 2
        else:
            cnt = int(cnt_str)
        name = re.sub(multi_ctrl_reg, "mc", name)
    return name, cnt



class DrawElement:
    """Draw Element."""

    def __init__(self, label=None):
        self._width = None
        self.label = self.mid_content = label
        self.top_format = self.mid_format = self.bot_format = "%s"
        self.top_connect = self.bot_connect = " "
        self.top_pad = self._mid_padding = self.bot_pad = " "
        self.mid_bck = self.top_bck = self.bot_bck = " "
        self.bot_connector = {}
        self.top_connector = {}
        self.right_fill = self.left_fill = self.layer_width = 0
        self.wire_label = ""

    @property
    def top(self):
        """Constructs the top line of the element."""
        if (
            (self.width % 2) == 0
            and len(self.top_format) % 2 == 1
            and len(self.top_connect) == 1
        ):
            ret = self.top_format % (self.top_pad + self.top_connect).center(
                self.width, self.top_pad
            )
        else:
            ret = self.top_format % self.top_connect.center(self.width, self.top_pad)
        if self.right_fill:
            ret = ret.ljust(self.right_fill, self.top_pad)
        if self.left_fill:
            ret = ret.rjust(self.left_fill, self.top_pad)
        ret = ret.center(self.layer_width, self.top_bck)
        return ret

    @property
    def mid(self):
        """Constructs the middle line of the element."""
        ret = self.mid_format % self.mid_content.center(self.width, self._mid_padding)
        if self.right_fill:
            ret = ret.ljust(self.right_fill, self._mid_padding)
        if self.left_fill:
            ret = ret.rjust(self.left_fill, self._mid_padding)
        ret = ret.center(self.layer_width, self.mid_bck)
        return ret

    @property
    def bot(self):
        """Constructs the bottom line of the element."""
        if (self.width % 2) == 0 and len(self.top_format) % 2 == 1:
            ret = self.bot_format % (self.bot_pad + self.bot_connect).center(
                self.width, self.bot_pad
            )
        else:
            ret = self.bot_format % self.bot_connect.center(self.width, self.bot_pad)
        if self.right_fill:
            ret = ret.ljust(self.right_fill, self.bot_pad)
        if self.left_fill:
            ret = ret.rjust(self.left_fill, self.bot_pad)
        ret = ret.center(self.layer_width, self.bot_bck)
        return ret

    @property
    def length(self):
        """Returns the length of the element, including the box around."""
        return max(len(self.top), len(self.mid), len(self.bot))

    @property
    def width(self):
        """Returns the width of the label, including padding."""
        if self._width:
            return self._width
        return len(self.mid_content)

    @width.setter
    def width(self, value):
        self._width = value

    def connect(self, wire_char, where, label=None):
        """Connects boxes and elements using wire_char and setting proper connectors.

        Args:
            wire_char (char): For example '║' or '│'.
            where (list["top", "bot"]): Where the connector should be set.
            label (string): Some connectors have a label (see cu1, for example).
        """

        if "top" in where and self.top_connector:
            self.top_connect = self.top_connector[wire_char]

        if "bot" in where and self.bot_connector:
            self.bot_connect = self.bot_connector[wire_char]

        if label:
            self.top_format = self.top_format[:-1] + (label if label else "")


class BoxOnClWire(DrawElement):
    """Draws a box on the classical wire.

    Args:
        label: Used as mid_content.
        top_connect: top_connect.
        bot_connect: bot_connect.

    Example:
    ::

        top: ┌───┐   ┌───┐
        mid: ╡ A ╞ ══╡ A ╞══
        bot: └───┘   └───┘
    """

    def __init__(self, label="", top_connect="─", bot_connect="─"):
        super().__init__(label)
        self.top_format = "┌─%s─┐"
        self.mid_format = "╡ %s ╞"
        self.bot_format = "└─%s─┘"
        self.top_pad = self.bot_pad = "─"
        self.mid_bck = "═"
        self.top_connect = top_connect
        self.bot_connect = bot_connect
        self.mid_content = label


class BoxOnQuWire(DrawElement):
    """Draws a box on the quantum wire.

    Example:
    ::

        top: ┌───┐   ┌───┐
        mid: ┤ A ├ ──┤ A ├──
        bot: └───┘   └───┘
    """

    def __init__(self, label="", top_connect="─", conditional=False):
        super().__init__(label)
        self.top_format = "┌─%s─┐"
        self.mid_format = "┤ %s ├"
        self.bot_format = "└─%s─┘"
        self.top_pad = self.bot_pad = self.mid_bck = "─"
        self.top_connect = top_connect
        self.bot_connect = "╥" if conditional else "─"
        self.mid_content = label
        self.top_connector = {"│": "┴"}
        self.bot_connector = {"│": "┬"}


class MeasureTo(DrawElement):
    """The element on the classic wire to which the measure is performed.

    Example:
    ::

        top:  ║     ║
        mid: ═╩═ ═══╩═══
        bot:
    """

    def __init__(self, label=""):
        super().__init__()
        self.top_connect = " ║ "
        self.mid_content = "═╩═"
        self.bot_connect = label
        self.mid_bck = "═"


class MeasureFrom(BoxOnQuWire):
    """The element on the quantum wire in which the measure is performed.

    Example:
    ::

        top: ┌─┐    ┌─┐
        mid: ┤M├ ───┤M├───
        bot: └╥┘    └╥┘
    """

    def __init__(self):
        super().__init__()
        self.top_format = self.mid_format = self.bot_format = "%s"
        self.top_connect = "┌─┐"
        self.mid_content = "┤M├"
        self.bot_connect = "└╥┘"

        self.top_pad = self.bot_pad = " "
        self._mid_padding = "─"


class MultiBox(DrawElement):
    """Elements that are drawn over multiple wires."""

    def center_label(self, input_length, order):
        """In multi-bit elements, the label is centered vertically."""
        if input_length == order == 0:
            self.top_connect = self.label
            return

        location_in_the_box = "*".center(input_length * 2 - 1).index("*") + 1
        top_limit = order * 2 + 2
        bot_limit = top_limit + 2

        if top_limit <= location_in_the_box < bot_limit:
            if location_in_the_box == top_limit:
                self.top_connect = self.label
            elif location_in_the_box == top_limit + 1:
                self.mid_content = self.label
            else:
                self.bot_connect = self.label

    @property
    def width(self):
        """Returns the width of the label, including padding."""
        if self._width:
            return self._width
        return len(self.label)


class BoxOnQuWireTop(MultiBox, BoxOnQuWire):
    """Draws the top part of a box that affects more than one quantum wire."""

    def __init__(self, label="", top_connect=None, wire_label=""):
        super().__init__(label)
        self.wire_label = wire_label
        self.bot_connect = self.bot_pad = " "
        self.mid_content = ""
        self.left_fill = len(self.wire_label)
        self.top_format = "┌─" + "s".center(self.left_fill + 1, "─") + "─┐"
        self.top_format = self.top_format.replace("s", "%s")
        self.mid_format = f"┤{self.wire_label} %s ├"
        self.bot_format = f"│{self.bot_pad * self.left_fill} %s │"
        self.top_connect = top_connect if top_connect else "─"


class BoxOnWireMid(MultiBox):
    """A generic middle box."""

    def __init__(self, label, input_length, order, wire_label=""):
        super().__init__(label)
        self.top_pad = self.bot_pad = self.top_connect = self.bot_connect = " "
        self.wire_label = wire_label
        self.left_fill = len(self.wire_label)
        self.top_format = f"│{self.top_pad * self.left_fill} %s │"
        self.bot_format = f"│{self.bot_pad * self.left_fill} %s │"
        self.top_connect = self.bot_connect = self.mid_content = ""
        self.center_label(input_length, order)


class BoxOnQuWireMid(BoxOnWireMid, BoxOnQuWire):
    """Draws the middle part of a box that affects more than one quantum wire."""

    def __init__(self, label, input_length, order, wire_label="", control_label=None):
        super().__init__(label, input_length, order, wire_label=wire_label)
        if control_label:
            self.mid_format = f"{control_label}{self.wire_label} %s ├"
        else:
            self.mid_format = f"┤{self.wire_label} %s ├"


class BoxOnQuWireBot(MultiBox, BoxOnQuWire):
    """Draws the bottom part of a box that affects more than one quantum wire."""

    def __init__(
        self, label, input_length, bot_connect=None, wire_label="", conditional=False
    ):
        super().__init__(label)
        self.wire_label = wire_label
        self.top_pad = " "
        self.left_fill = len(self.wire_label)
        self.top_format = f"│{self.top_pad * self.left_fill} %s │"
        self.mid_format = f"┤{self.wire_label} %s ├"
        self.bot_format = "└─" + "s".center(self.left_fill + 1, "─") + "─┘"
        self.bot_format = self.bot_format.replace("s", "%s")
        bot_connect = bot_connect if bot_connect else "─"
        self.bot_connect = "╥" if conditional else bot_connect

        self.mid_content = self.top_connect = ""
        if input_length <= 2:
            self.top_connect = label


class BoxOnClWireTop(MultiBox, BoxOnClWire):
    """Draws the top part of a conditional box that affects more than one classical wire."""

    def __init__(self, label="", top_connect=None, wire_label=""):
        super().__init__(label)
        self.wire_label = wire_label
        self.mid_content = ""
        self.bot_format = "│ %s │"
        self.top_connect = top_connect if top_connect else "─"
        self.bot_connect = self.bot_pad = " "


class BoxOnClWireMid(BoxOnWireMid, BoxOnClWire):
    """Draws the middle part of a conditional box that affects more than one classical wire."""

    def __init__(self, label, input_length, order, wire_label="", **_):
        super().__init__(label, input_length, order, wire_label=wire_label)
        self.mid_format = f"╡{self.wire_label} %s ╞"


class BoxOnClWireBot(MultiBox, BoxOnClWire):
    """Draws the bottom part of a conditional box that affects more than one classical wire."""

    def __init__(self, label, input_length, bot_connect="─", wire_label="", **_):
        super().__init__(label)
        self.wire_label = wire_label
        self.left_fill = len(self.wire_label)
        self.top_pad = " "
        self.bot_pad = "─"
        self.top_format = f"│{self.top_pad * self.left_fill} %s │"
        self.mid_format = f"╡{self.wire_label} %s ╞"
        self.bot_format = "└─" + "s".center(self.left_fill + 1, "─") + "─┘"
        self.bot_format = self.bot_format.replace("s", "%s")
        bot_connect = bot_connect if bot_connect else "─"
        self.bot_connect = bot_connect

        self.mid_content = self.top_connect = ""
        if input_length <= 2:
            self.top_connect = label


class DirectOnQuWire(DrawElement):
    """Element to the wire (without the box)."""

    def __init__(self, label=""):
        super().__init__(label)
        self.top_format = " %s "
        self.mid_format = "─%s─"
        self.bot_format = " %s "
        self._mid_padding = self.mid_bck = "─"
        self.top_connector = {"│": "│"}
        self.bot_connector = {"│": "│"}


class Barrier(DirectOnQuWire):
    """Draws a barrier.

    Example:

        top:  ░     ░
        mid: ─░─ ───░───
        bot:  ░     ░
    """

    def __init__(self, label=""):
        super().__init__("░")
        self.top_connect = "░"
        self.bot_connect = "░"
        self.top_connector = {}
        self.bot_connector = {}


class Ex(DirectOnQuWire):
    """Draws an X (usually with a connector).

    Example:
        top:
        mid: ─X─ ───X───
        bot:  │     │
    """

    def __init__(self, bot_connect=" ", top_connect=" ", conditional=False):
        super().__init__("X")
        self.bot_connect = "║" if conditional else bot_connect
        self.top_connect = top_connect


class iEx(DirectOnQuWire):
    """Draws an ⨂ (usually with a connector).

    Example:

        top:
        mid: ─⨂─ ───⨂───
        bot:  │      │
    """

    def __init__(self, bot_connect="  ", top_connect=" ", conditional=False):
        super().__init__("⨂")
        self.bot_connect = "║" if conditional else bot_connect
        self.top_connect = top_connect


class ResetDisplay(DirectOnQuWire):
    """Draws a reset gate."""

    def __init__(self, conditional=False):
        super().__init__("|0>")
        if conditional:
            self.bot_connect = "║"


class Bullet(DirectOnQuWire):
    """Draws a bullet (usually with a connector).

    Example:
        top:
        mid: ─■─  ───■───
        bot:  │      │
    """

    def __init__(
        self,
        top_connect="",
        bot_connect="",
        conditional=False,
        label=None,
        bottom=False,
    ):
        super().__init__("■")
        self.conditional = conditional
        self.top_connect = top_connect
        self.bot_connect = "║" if conditional else bot_connect
        if label and bottom:
            self.bot_connect = label
        elif label:
            self.top_connect = label
        self.mid_bck = "─"


class OpenBullet(DirectOnQuWire):
    """Draws an open bullet (usually with a connector).

    Example:
        top:
        mid: ─o─  ───o───
        bot:  │      │
    """

    def __init__(
        self,
        top_connect="",
        bot_connect="",
        conditional=False,
        label=None,
        bottom=False,
    ):
        super().__init__("o")
        self.conditional = conditional
        self.top_connect = top_connect
        self.bot_connect = "║" if conditional else bot_connect
        if label and bottom:
            self.bot_connect = label
        elif label:
            self.top_connect = label
        self.mid_bck = "─"


class DirectOnClWire(DrawElement):
    """Element to the classical wire (without the box)."""

    def __init__(self, label=""):
        super().__init__(label)
        self.top_format = " %s "
        self.mid_format = "═%s═"
        self.bot_format = " %s "
        self._mid_padding = self.mid_bck = "═"
        self.top_connector = {"│": "│"}
        self.bot_connector = {"│": "│"}


class ClBullet(DirectOnClWire):
    """Draws a bullet on classical wire (usually with a connector).

    Example:
        top:
        mid: ═■═  ═══■═══
        bot:  │      │
    """

    def __init__(
        self,
        top_connect="",
        bot_connect="",
        conditional=False,
        label=None,
        bottom=False,
    ):
        super().__init__("■")
        self.top_connect = top_connect
        self.bot_connect = "║" if conditional else bot_connect
        if label and bottom:
            self.bot_connect = label
        elif label:
            self.top_connect = label
        self.mid_bck = "═"


class ClOpenBullet(DirectOnClWire):
    """Draws an open bullet on classical wire (usually with a connector).

    Example:
        top:
        mid: ═o═  ═══o═══
        bot:  │      │
    """

    def __init__(
        self,
        top_connect="",
        bot_connect="",
        conditional=False,
        label=None,
        bottom=False,
    ):
        super().__init__("o")
        self.top_connect = top_connect
        self.bot_connect = "║" if conditional else bot_connect
        if label and bottom:
            self.bot_connect = label
        elif label:
            self.top_connect = label
        self.mid_bck = "═"


class EmptyWire(DrawElement):
    """This element is just the wire, with no operations."""

    def __init__(self, wire):
        super().__init__(wire)
        self._mid_padding = self.mid_bck = wire

    @staticmethod
    def fillup_layer(layer, first_clbit):
        """Given a layer, replace the Nones in it with EmptyWire elements."""
        for nones in [i for i, x in enumerate(layer) if x is None]:
            layer[nones] = EmptyWire("═") if nones >= first_clbit else EmptyWire("─")
        return layer


class BreakWire(DrawElement):
    """This element is used to break the drawing in several pages."""

    def __init__(self, arrow_char):
        super().__init__()
        self.top_format = self.mid_format = self.bot_format = "%s"
        self.top_connect = arrow_char
        self.mid_content = arrow_char
        self.bot_connect = arrow_char

    @staticmethod
    def fillup_layer(layer_length, arrow_char):
        """Creates a layer with BreakWire elements."""
        breakwire_layer = []
        for _ in range(layer_length):
            breakwire_layer.append(BreakWire(arrow_char))
        return breakwire_layer


class InputWire(DrawElement):
    """This element is the label and the initial value of a wire."""

    def __init__(self, label):
        super().__init__(label)

    @staticmethod
    def fillup_layer(names):
        """Creates a layer with InputWire elements."""
        longest = max(len(name) for name in names)
        inputs_wires = []
        for name in names:
            inputs_wires.append(InputWire(name.rjust(longest)))
        return inputs_wires



from typing import Union

from qutrunk.circuit import Qureg
from qutrunk.circuit.ops import Operator
from qutrunk.exceptions import QuTrunkError
from qutrunk.circuit.gates import X


class Classical(Operator):
    """Quantum state preparation Operator.

    Init the quantum state to classical state.

    Args:
        state: The target state of circuit initialized to.

    Example:
        .. code-block:: python

            from qutrunk.circuit.ops import Classical
            from qutrunk.circuit import QCircuit
            from qutrunk.circuit.gates import H, All, Measure

            circuit = QCircuit()
            qureg = circuit.allocate(3)
            Classical("0100") * qureg 
            # Classical(4) * qureg
            print(circuit.get_statevector())
    """

    def __init__(self, state: Union[str, int]):
        super().__init__()
        self.state = state

    def __str__(self):
        return "Classical"

    def __mul__(self, qureg: Qureg):
        if qureg.circuit.num_gates > 0:
            raise QuTrunkError("Classical operator should be applied at the beginning of circuit.")

        if not isinstance(qureg, Qureg):
            raise TypeError("the operand must be Qureg.")

        if not self._check_state(qureg):
            raise ValueError(f"Invalid state: {self.state}")

        self._process_state(qureg)

    def _check_state(self, qureg: Qureg):
        if isinstance(self.state, str):
            val = int(self.state, base=2)
            if 0 <= val < 2 ** len(qureg):
                return True

        if isinstance(self.state, int):
            if 0 <= self.state < 2 ** len(qureg):
                return True

        return False

    def _process_state(self, qureg: Qureg):
        bit_strs = ""
        if isinstance(self.state, str):
            if self.state.startswith("0b"):
                bit_strs = self.state[2:]
            else:
                bit_strs = self.state
        elif isinstance(self.state, int):
            bit_strs = str(bin(self.state))
            bit_strs = bit_strs[2:]

        bit_strs = bit_strs.zfill(len(qureg))
        bit_strs = bit_strs[::-1]

        for i, _ in enumerate(qureg):
            if bit_strs[i] == "1":
                X * qureg[i]



"""AWS Braket backends."""

import datetime
from abc import ABC

from braket.aws import AwsDevice
from braket.circuits import Circuit
from braket.device_schema.dwave import DwaveDeviceCapabilities
from braket.device_schema.xanadu import XanaduDeviceCapabilities
from braket.devices import LocalSimulator
from braket.tasks.local_quantum_task import LocalQuantumTask

from .adapter import convert_qutrunk_to_braket_circuit
from .braket_job import AWSBraketJob
from .exception import QuTrunkBraketException
from .. import Backend

CLIENT_VERSION = "0.0.1"


class BackendBraket(Backend, ABC):
    """BraketBackend."""

    def __init__(
            self,
            device,
            name: str = None,
            description: str = None,
            online_date: datetime.datetime = None,
    ):
        """Initialize a based backend

        Args:
            device: device: Braket device class
            online_date:  online date
            name: An optional name for the backend
            description: An optional description of the backend

        Raises:
            AttributeError: If a field is specified that's outside the backend's
                options
        """
        self._device = device
        self._name = name
        self._description = description
        self._online_date = online_date
        self._status = device.status
        self._circuit = None

    def send_circuit(self, circuit, final=False):
        self._circuit = circuit

    def run(self, shots=1024):
        circuit: Circuit = convert_qutrunk_to_braket_circuit(self._circuit)
        try:
            task: LocalQuantumTask = self._device.run(
                task_specification=circuit, shots=shots
            )
        except Exception as ex:
            task.cancel()
            raise ex

        return AWSBraketJob(
            job_id=task.id,
            backend=self,
            task=task,
            shots=shots,
        ).result()

    @property
    def name(self):
        return f"BackendAWSBraket[{self._name}]"


class BackendAWSLocal(BackendBraket):
    """BraketLocalBackend."""

    def __init__(self, name: str = "default"):
        """AWSBraketLocalBackend for local execution of circuits.

        Example:
            .. code-block:: python

                device = LocalSimulator()                         #Local State Vector Simulator
                device = LocalSimulator("default")                #Local State Vector Simulator
                device = LocalSimulator(name="default")        #Local State Vector Simulator
                device = LocalSimulator(name="braket_sv")      #Local State Vector Simulator
                device = LocalSimulator(name="braket_dm")      #Local Density Matrix Simulator

        Args:
            name: name of backend
        """
        device = LocalSimulator(backend=name)
        super().__init__(device, name="sv_simulator")


class BackendAWSDevice(BackendBraket):
    """ AWSBraketBackend.
        following below steps to use AWS braket device.
        referring to https://docs.aws.amazon.com/braket/index.html for detail
        1. sign in AWS console, visit https://console.aws.amazon.com/braket/home.
            follow guide to create execution roles, including service-link role and job execute role
        2. set up AWS credentials and default region,
            referring to https://boto3.amazonaws.com/v1/documentation/api/latest/guide/quickstart.html
    """

    def __init__(self, name: str = None):
        """AWSBraketBackend for execution circuits against AWS Braket devices.

        Args:
            name: name of backend, such as SV1, TN1, DM1, etc. refer to Amazon Braket service for detail
        """
        device = supported_aws_device(name)[0]
        user_agent = f"QuTrunkBraketProvider/{CLIENT_VERSION}"
        device.aws_session.add_braket_user_agent(user_agent)
        super().__init__(
            device,
            name=name,
            description=f"AWS Device: {device.provider_name} {device.name}.",
            online_date=device.properties.service.updatedAt,
        )


def supported_aws_device(name):
    """

    Args:
        name: device name, such as
         [Aspen-10],
         [Aspen-11],
         [Aspen-8],
         [Aspen-9],
         [Aspen-M-1],
         [IonQ Device],
         [Lucy],
         [SV1],
         [TN1],
         [dm1]]

    Returns:
        supported device, such as
         [BraketBackend[Aspen-10],
         BraketBackend[Aspen-11],
         BraketBackend[Aspen-8],
         BraketBackend[Aspen-9],
         BraketBackend[Aspen-M-1],
         BraketBackend[IonQ Device],
         BraketBackend[Lucy],
         BraketBackend[SV1],
         BraketBackend[TN1],
         BraketBackend[dm1]]

    """
    names = [name] if name else None
    devices = AwsDevice.get_devices(names=names)
    supported_devices = [
        d
        for d in devices
        if not isinstance(
            d.properties, (DwaveDeviceCapabilities, XanaduDeviceCapabilities)
        )
    ]

    if not supported_devices:
        raise QuTrunkBraketException(f"no available aws device for '{name}'")
    return supported_devices
